#
# -*- coding: utf-8
# vim: set fileencoding=utf-8

# This code is Copyright (c) 2008 Thomas Capricelli <orzel@freehackers.org>
# distributed under the GNU General Public Licence GPL, version 2

# The homepage is http://www.freehackers.org/GentooDepBrowser
# There, you should find latest version, documentation and such.


import os, sys, time
from PyQt4 import Qt, QtGui, QtCore
from stat import ST_MTIME
import pickle


#
# config stuff
#
CONFIG = {}
# if true, resolve dependencies (using USE for exemple), else consider ALL
# dependencies
CONFIG["resolveDepStrings"] = True
CONFIG["stripAtoms"] = True
CONFIG["onlyInstalled"] = True
CONFIG["cacheFile"] = 'GentooDepBrowser.cache.pkl'

#
#portage stuff
#
import portage, portage_dep, portage_util
PORTAGE_SETTINGS = portage.config(clone = portage.settings)

#try:
#    from metalib import *
#except ImportError:
#    sys.stderr.write("Could not import metalib.py. Maybe you forgot to download it?")
#    sys.exit(1)


DB = portage.db["/"]["porttree"].dbapi # use /usr/portage
#DB = portage.db["/"]["vartree"].dbapi # use /var/deb/..

def resolve_dep_strings(cpv, values):
    PORTAGE_SETTINGS.setcpv(cpv)
    result = portage.flatten(portage_dep.use_reduce(portage_dep.paren_reduce(values), PORTAGE_SETTINGS["USE"].split()))
    return result

def strip_atoms(parts):
    result = []
    for x in parts:
        if portage.isvalidatom(x):
            result.append(portage.dep_getkey(x))
        else:
            result.append(x)
    return result

def portageGetDependencies(cpv):
    keys = ['DEPEND']
    try:
        values = DB.aux_get(cpv, keys)
    except KeyError:
        print "KeyError while fetching dependencies for %s" % cpv
        return []
    if values == ['']: return []
    # remove conditional deps whose coditions aren't met
    if CONFIG["resolveDepStrings"]:
        values = resolve_dep_strings(cpv, values[0])
    values = filter(lambda a: not a.find('||')>=0 , values)
    # user requested only package names
    if CONFIG["stripAtoms"]:
        values = strip_atoms(values)
    return filter(lambda a: not a.startswith('!'), values)

#
# Db stuff
#

def removeDuplicate(seq): 
    keys = {} 
    for e in seq: 
        keys[e] = 1 
    return list(keys)


class PackageDict:
    def __init__(self):
        self.initialized = False
        self.data = {}

    def load(self, gui = False):
        # cache ?
        content = portage_util.grabfile( os.path.join( PORTAGE_SETTINGS["PORTDIR"], "metadata", "timestamp.chk"))
        portagetimestamp = 10
        if content:
            try:
                portagetimestamp = time.mktime(time.strptime(content[0], "%a, %d %b %Y %H:%M:%S +0000"))
            except (OverflowError, ValueError):
                pass
        del content
        filetimestamp =0
        try:
            filetimestamp = os.stat(CONFIG["cacheFile"])[ST_MTIME]
        except OSError:
            pass
        if filetimestamp >portagetimestamp:
            self.loadFromDisk()
        # if for some reason self.loadFromDisk() failed:
        if not self.initialized:
            self.constructFromPortage(gui)

    def constructFromPortage(self, gui):
        # read DB
        if gui:
            
            dialog = QtGui.QDialog()
            dialog.setWindowTitle("Building the package database")
            dialog.setMinimumWidth(500)
            layout = QtGui.QGridLayout()
            layout.addWidget( QtGui.QLabel("Parse portage databse"))
            qpb1 = QtGui.QProgressBar()
            layout.addWidget( qpb1)
            layout.addWidget( QtGui.QLabel("Create dependencies tree"))
            qpb2 = QtGui.QProgressBar()
            layout.addWidget( qpb2)
            layout.addWidget( QtGui.QLabel("Remove duplicates"))
            qpb3 = QtGui.QProgressBar()
            layout.addWidget( qpb3)
            dialog.setLayout(layout)
            dialog.show()
            Qt.qApp.processEvents()
        packages = DB.cp_all()
        nb = len(packages)
        if gui:
            qpb1.setMaximum(nb)
            qpb2.setMaximum(nb)
            qpb3.setMaximum(nb)
        i = 0
        for name in packages:
            # the package will register itself is relevant
            p = Package(name)
            if p.cpv!='':
                self.data[name] = p
            i+=1
            if gui:
                qpb1.setValue(i)
                Qt.qApp.processEvents()
        # create dependencies tree
        nb = len(self.data)
        i = 0
        for name, package in self.data.items():
            package.constructDepTree(self.data)
            i = i+1
            if gui:
                qpb2.setValue(i)
                Qt.qApp.processEvents()
        # remove duplicates
        i = 0
        for name, package in self.data.items():
            package.dependencies = removeDuplicate(package.dependencies)
            package.reverseDependencies = removeDuplicate(package.reverseDependencies)
            i = i+1
            if gui:
                qpb3.setValue(i)
                Qt.qApp.processEvents()
        # cache it
        self.initialized = True
        self.saveToDisk()
    def saveToDisk(self):
        if not self.initialized:
            print "PackageDict.saveToDisk called while not initialized, aborting"
            return;
        output = open(CONFIG["cacheFile"], 'wb')
        pickle.dump(self.data, output)
        output.close()
    def loadFromDisk(self):
        if self.initialized:
            print "PackageDict.loadFromDisk while self.initialized is True, strange"
            #return # already done
        output = open(CONFIG["cacheFile"], 'r')
        self.data =  pickle.load(output)
        #print "loadFromdisk loaded %d elements"% len(self.data)
        output.close()
        self.initialized = True
    def asStringList(self):
        return list(self.data)

class Package:
    "This class stores information about a package"
    def __init__(self, name):
        self.name = name
        self.dependencies = []
        self.reverseDependencies = []
        # get dep information from portage
        if not portage.isvalidatom(self.name):
            raise ValueError
        self.cpv = portage.best(DB.match(self.name))
        if self.cpv == "":
            #print "cant find best package for name=%s" % (self.name)
            return
    def __unicode__(self):
        s  = "\n================"
        s += "\npackage name : %s" % self.name
        s += "\ndependencies : "
        s += str([p.name for p in self.dependencies])
        s += "\nReverse dep  : "
        s += str([p.name for p in self.reverseDependencies])
        return s
    def __str__(self):
        return self.__unicode__()
    def constructDepTree(self, packageDict):
        dependenciesStr = portageGetDependencies(self.cpv)
        # register this
        for pname in dependenciesStr:
            if pname.startswith('virtual'):
                continue
            if not packageDict.has_key(pname):
                #print "%s : error looking for package %s indict" % (self.name, pname)
                continue
            package = packageDict[pname]
            self.dependencies.append(package) #forward
            package.reverseDependencies.append(self) # backward

#
# GUI Stuff
#
class PackageNode(QtGui.QGraphicsRectItem):
    def __init__(self, name, color = QtCore.Qt.red):
        QtGui.QGraphicsRectItem.__init__(self)
        self.text = QtGui.QGraphicsTextItem(name, self)
        rect = self.text.boundingRect()
        self.text.moveBy(-rect.center().x(),-rect.center().y())
        rect.translate(-rect.center())
        self.setRect( rect.adjusted(-3,-3,3,3))
        self.setBrush(QtCore.Qt.white)
        self.setPen(color)
        self.setZValue(-10)
        self.text.setZValue(10)
        # print "creating new PackageNode"
    def mousePressEvent(self, event):
        global globalview
        package = unicode(self.text.toPlainText())
        globalview.userClick(package)

class MyView(QtGui.QGraphicsView):
    def __init__(self):
        QtGui.QGraphicsView.__init__(self)

        self.scene = QtGui.QGraphicsScene(QtCore.QRectF(0,0,1000,6000), self)
        self.resize(1050,850)
        self.setScene(self.scene)
        self.history = []
        self.history_pos = 0
#        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
#        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)

        self.packageDict = PackageDict()

    #def resizeEvent(event):
    def displayPackageList(self, plist, displayOnTop = True):
        import copy
        plist = copy.copy(plist) # deep copy, we are going to destroy it
        rect = self.scene.sceneRect()
        plistlen = len(plist)
        if plistlen <= 0:return

        # environement
        if displayOnTop:
            dy = -50
        else:
            dy = 50
        x, y = 0, rect.center().y()+dy

        # actual display
        # pack as much as possible in a row
        if len(plist)>7:
            while len(plist)>0:
                package = plist.pop()
                if displayOnTop:
                    packageNode = PackageNode(package.name, QtCore.Qt.red)
                else:
                    packageNode = PackageNode(package.name, QtCore.Qt.green)
                nextx = x + packageNode.boundingRect().width()+3
                if nextx > rect.right():
                    if len(plist)<8:
                        del packageNode
                        plist.append(package)
                        break;
                    x, nextx = 0, packageNode.boundingRect().width()+3
                    y+= dy
                packageNode.setPos(x+packageNode.boundingRect().width()/2, y)
                self.scene.addItem(packageNode)
                link = QtGui.QGraphicsLineItem(QtCore.QLineF(packageNode.pos(), rect.center()))
                link.setZValue(-20)
                self.scene.addItem(link)
                x = nextx
        # more space between them
        if len(plist)<=0: return
        dx = rect.width()/len(plist)
        x = dx/2
        y += dy
        while len(plist)>0:
            package = plist.pop()
            if displayOnTop:
                packageNode = PackageNode(package.name, QtCore.Qt.red)
            else:
                packageNode = PackageNode(package.name, QtCore.Qt.green)
            packageNode.setPos(x, y)
            self.scene.addItem(packageNode)
            link = QtGui.QGraphicsLineItem(QtCore.QLineF(packageNode.pos(), rect.center()))
            link.setZValue(-20)
            self.scene.addItem(link)
            x += dx

    def cleanScene(self):
        for element in self.scene.items():
            if element.scene():
                #print "deleting ", element
                self.scene.removeItem(element)
            del(element)
        #print "cleanScene : %d elements left" % len(self.scene.items())

    def cleanPackageName(self, name):
        #print "cleanPackageName for "+name
        if name.startswith(('=','<','>','!')):
            name = name.lstrip('=><!')
            name = '-'.join(name.split('-')[:-1])
        #print "returning "+name
        return name

    def userClick(self,package):
        # cut history tail
        if len(self.history)-1 > self.history_pos:
            self.history=self.history[:self.history_pos+1]
        # add the package
        self.history.append(package)
        self.history_pos=len(self.history)-1
        #print "userClick, history is ", self.history
        #print "userClick, position is ",self.history_pos
        self.displayOnePackage(package)

    def goToPrevious(self):
        if self.history_pos>0:
            self.history_pos-=1
            self.displayOnePackage(self.history[self.history_pos])

    def goToNext(self):
        if len(self.history)-1 > self.history_pos:
            self.history_pos+=1
            self.displayOnePackage(self.history[self.history_pos])

    def displayOnePackage(self, package):
        # check package
        if not portage.isvalidatom(package):
            raise ValueError
        if not self.packageDict.data.has_key(package):
            raise ValueError

        # prepare display
        self.cleanScene()
        rect = self.scene.sceneRect()
        mainpackage = PackageNode(package, QtCore.Qt.black)
        mainpackage.setPos(rect.center())
        self.scene.addItem(mainpackage)
        self.centerOn(rect.center())

        # package depends on those pacckages
        plist = self.packageDict.data[package].dependencies
        self.displayPackageList(plist, False)

        # Packages that depend on 'package'
        plist = self.packageDict.data[package].reverseDependencies
        self.displayPackageList(plist, True)


class SearchDialog(QtGui.QDialog):
    def __init__(self, parent = None):
        global globalview
        QtGui.QDialog.__init__(self, parent)
        
        self.setMinimumWidth(430)
        self.setMinimumHeight(600)

        lineEdit = QtGui.QLineEdit()
        lineEdit.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Preferred)

        self.listBox = QtGui.QListWidget()

        self.packagesName = globalview.packageDict.asStringList()
        self.listBox.addItems(QtCore.QStringList(self.packagesName))
    
        cancelButton = QtGui.QPushButton(self.tr("Cancel"))
    
        self.connect(lineEdit, QtCore.SIGNAL("textEdited(const QString&)"), self.updateList)
        self.connect(self.listBox, QtCore.SIGNAL("itemActivated (QListWidgetItem *)"), self.itemActivated)
        self.connect(cancelButton, QtCore.SIGNAL("clicked()"), self, QtCore.SLOT("reject()"))
    
        buttonLayout = QtGui.QHBoxLayout()
        buttonLayout.addStretch(1)
        buttonLayout.addWidget(cancelButton)
    
        mainLayout = QtGui.QVBoxLayout()
        mainLayout.addWidget(lineEdit)
        mainLayout.addWidget(self.listBox)
        mainLayout.addLayout(buttonLayout)
        self.setLayout(mainLayout)
    
        self.setWindowTitle(self.tr("Search package"))
    def updateList(self, text):
        self.listBox.clear()
        for n in self.packagesName:
            if n.find(text)>=0:
                self.listBox.addItem(n)
    def itemActivated(self, item):
        global globalview
        self.accept()
        globalview.userClick(unicode(item.text()))
        
class MainWindow(QtGui.QMainWindow):
    def __init__(self):
        global globalview, globalmainwindow
        QtGui.QMainWindow.__init__(self)

        self.setMinimumWidth(600)
        self.setMinimumHeight(400)

        self.view =  MyView()
        globalview = self.view
        globalmainwindow = self
        self.setCentralWidget(self.view)

        self.createActions()
        self.createMenus()
        self.createToolBars()
        self.createStatusBar()

        self.readSettings()

        self.show()
        self.view.packageDict.load(True)
        self.view.userClick('media-libs/mesa')

    def rescan(self):
        self.view.packageDict.constructFromPortage(True)

    def openFindDialog(self):
        dialog = SearchDialog()
        dialog.exec_()

    def closeEvent(self, event):
        self.writeSettings()
        event.accept()

    def about(self):
        QtGui.QMessageBox.about(self, self.tr("About GentooDepBrowser"),
            self.tr("<b>GentooDepBrowser</b> is a graphic tool to help understand package "
            "dependancies in gentoo."
            "You can browse dependencies by clicking on packages names."
            "<br/>"
            "<br/>"
            "Copyright (c) 2008 Thomas Capricelli"
            "<br/>"
            '<a href="http://www.freehackers.org/GentooDepBrowser">http://www.freehackers.org/GentooDepBrowser</a>'
            "<br/>"
            "<br/>"
            '<a href="http://www.freehackers.org/bugtracker/set_project.php?project_id=12&ref=bug_report_page.php">Report a bug or wish</a>'
            ))

    def createActions(self):
        self.rescanAct = QtGui.QAction(self.tr("&Rescan..."), self)
        self.rescanAct.setStatusTip(self.tr("Parse the whole portage tree again"))
        self.connect(self.rescanAct, QtCore.SIGNAL("triggered()"), self.rescan)

        self.exitAct = QtGui.QAction(self.tr("E&xit"), self)
        self.exitAct.setShortcut(self.tr("Ctrl+Q"))
        self.exitAct.setStatusTip(self.tr("Exit the application"))
        self.connect(self.exitAct, QtCore.SIGNAL("triggered()"), self, QtCore.SLOT("close()"))

        self.findAct = QtGui.QAction(self.tr("&Find package..."), self)
        self.findAct.setShortcut(self.tr("Ctrl+F"))
        self.findAct.setStatusTip(self.tr("Open a dialogs to find a package"))
        self.connect(self.findAct, QtCore.SIGNAL("triggered()"), self.openFindDialog)

        self.previousAct = QtGui.QAction(self.tr("&Previous"), self)
        self.previousAct.setStatusTip(self.tr("Go to previous package"))
        self.connect(self.previousAct, QtCore.SIGNAL("triggered()"), self.view.goToPrevious)

        self.nextAct = QtGui.QAction(self.tr("&Next"), self)
        self.nextAct.setStatusTip(self.tr("Go to next package"))
        self.connect(self.nextAct, QtCore.SIGNAL("triggered()"), self.view.goToNext)

        self.aboutAct = QtGui.QAction(self.tr("&About"), self)
        self.aboutAct.setStatusTip(self.tr("Show the application's About box"))
        self.connect(self.aboutAct, QtCore.SIGNAL("triggered()"), self.about)

    def createMenus(self):
        self.fileMenu = self.menuBar().addMenu(self.tr("&File"))
        self.fileMenu.addAction(self.rescanAct)
        self.fileMenu.addSeparator();
        self.fileMenu.addAction(self.exitAct)

        self.editMenu = self.menuBar().addMenu(self.tr("&Edit"))
        self.editMenu.addAction(self.findAct)
        self.editMenu.addAction(self.previousAct)
        self.editMenu.addAction(self.nextAct)

        self.menuBar().addSeparator()

        self.helpMenu = self.menuBar().addMenu(self.tr("&Help"))
        self.helpMenu.addAction(self.aboutAct)

    def createToolBars(self):
        self.editToolBar = self.addToolBar(self.tr("Edit"))
        self.editToolBar.addAction(self.rescanAct)
        self.editToolBar.addAction(self.findAct)
        self.editToolBar.addAction(self.previousAct)
        self.editToolBar.addAction(self.nextAct)

    def createStatusBar(self):
        self.statusBar().showMessage(self.tr("Ready"))

    def readSettings(self):
        settings = QtCore.QSettings("Freehackers", "GentooDepBrowser")
        pos = settings.value("pos", QtCore.QVariant(QtCore.QPoint(20, 20))).toPoint()
        size = settings.value("size", QtCore.QVariant(QtCore.QSize(1100, 900))).toSize()
        self.resize(size)
        self.move(pos)

    def writeSettings(self):
        settings = QtCore.QSettings("Freehackers", "GentooDepBrowser")
        settings.setValue("pos", QtCore.QVariant(self.pos()))
        settings.setValue("size", QtCore.QVariant(self.size()))


if __name__ == '__main__':
    sys.setrecursionlimit(30000)
    app = QtGui.QApplication(sys.argv)
    mainWin = MainWindow()
    mainWin.show()
    sys.exit(app.exec_())

# vim: ai ts=4 sts=4 et sw=4
